open Lber
open Ldap_types
open Ldap_ooclient
open Ldap_funserver
open Ldap_schemaparser
open Ldap_error
open Tdir_config
open Tdir_types
open Syslog
open Arg

let main () =
  let run_in_background = ref false in
  let config_location = ref "./tdir.conf" in
  let options = [("-d", Set run_in_background, "fork into the background and daemonize");
		 ("-f", Set_string config_location, "the path to the config file")]
  in
  let usage = "tdir [options]" in
  let _ = Arg.parse options (fun x -> raise (Invalid_argument x)) usage in
  let cfg = 
    try ref (getconfig !config_location) 
    with LDAP_Failure (c, e, ex) ->
      failwith (ldap_strerror "could not connect to ldap server" (LDAP_Failure (c, e, ex)))
  in
  let connections = 
    try
      {ad_schema=(let s = !cfg.cf_read_con#schema in Tdir_utils.alter_schema cfg s;s);
       connections=Hashtbl.create 50}
    with LDAP_Failure (c, e, ex) ->
      failwith (ldap_strerror "could not fetch schema" (LDAP_Failure (c, e, ex)))
  in
  let bi = 
    {bi_op_bind=Some (Tdir_bind.bind cfg connections);
     bi_op_unbind=Some (Tdir_unbind.unbind cfg connections);
     bi_op_search=None;
     bi_op_compare=None;
     bi_op_modify=Some (Tdir_modify.modify cfg connections);
     bi_op_modrdn=Some (Tdir_modrdn.modrdn cfg connections);
     bi_op_add=Some (Tdir_add.add cfg connections);
     bi_op_delete=Some (Tdir_delete.delete cfg connections);
     bi_op_abandon=None;
     bi_op_extended=None;
     bi_init=None;
     bi_close=None}
  in
  let start_server () = 
    Ldap_funserver.init 
      ~log:(fun _ msg -> syslog !cfg.cf_syslog `LOG_NOTICE msg)
      ~port:!cfg.cf_port bi 
  in
    (if !run_in_background then
       if (Unix.fork ()) = 0 then
	 (close_in stdin;
	  close_out stdout;
	  close_out stderr)
       else
	 exit 0);

    (* ignore broken pipes, which result from dead connections *)
    Sys.set_signal Sys.sigpipe Sys.Signal_ignore;
    try
      while true
      do      
	let si = start_server () in
	  try Ldap_funserver.run si;failwith "error! Ldap_funserver.run returned cleanly!"
	  with exn ->
	    syslog !cfg.cf_syslog `LOG_ERR
	      ("unhandled exception: " ^ (Printexc.to_string exn));
	    shutdown si;
	    Unix.sleep 1 (* prevent spinning error conditions *)
      done
    with
	Unix.Unix_error (e, f, a) ->
	  syslog !cfg.cf_syslog `LOG_ERR
	    ("fatal unix error: " ^ (Unix.error_message e) ^ " in: " ^ f ^ " with arg: " ^ a);
	  exit 2
      | exn ->
	  syslog !cfg.cf_syslog `LOG_ERR
	    ("fatal unhandled exception: " ^ (Printexc.to_string exn) ^ " exiting");
	  exit 3      
;;

main ()
