open Tdir_types
open Ldap_types
open Ldap_ooclient
open Ldap_schemaparser
open Ldap_error
open Str
open Syslog
open Printf

let lc = String.lowercase

let mod_res ?(referral=None) ?(matched_dn="") ?(result_code=`SUCCESS) 
            ?(error_msg="success") message =
  {messageID=message.messageID;
   protocolOp=Modify_response {result_code=result_code;
			       matched_dn=matched_dn;
			       error_message=error_msg;
			       ldap_referral=referral};
   controls=None}

let translate_modop cfg schema {mod_dn=dn;modification=mods} =
  let ad_dn = Tdir_utils.translate_dn cfg dn in
    (* we can't emit mods to userpassword directly for a few reasons,
       first AD doesn't use userpassword, it uses unicodepwd. Second,
       it uses the old NT SAM "sort of unicode" password format. We
       must therefore intercept all modifications to userpassword, and
       translate them into modifications of unicodepwd in the proper
       "sort of unicode" format. *)
  let translate_userpassword_mods_for_ad mods =
    let (userpassword_mods, other_mods) = (* seperate password mods from other mods *)
      List.fold_left
	(fun (userpassword_mods, other_mods) m -> 
	   match m with
	       {mod_value={attr_type=name}} when (lc name) = "userpassword" ->
		 (m :: userpassword_mods, other_mods)
	     | _ -> (userpassword_mods, m :: other_mods))
	([], [])
	mods
    in
      (List.rev_append (* put altered password mods back together with the other mods *)
	 other_mods
	 (List.flatten (* deal with all userpassword mods *)
	    (List.rev_map (* convert password ops to unicodePwd ops *)
	       (function
		    {mod_op=optype;
		     mod_value={attr_vals=attr_vals}}
		      when (optype = `ADD || optype = `REPLACE) ->
			[{mod_op=`REPLACE;
			  mod_value={attr_type="unicodepwd";
				     attr_vals=(List.rev_map 
						  Tdir_utils.mkunicodepwd 
						  attr_vals)}};
			 {mod_op=`REPLACE;
			  mod_value={attr_type="useraccountcontrol";
				     attr_vals=["66080"]}}]
		  | {mod_op=`DELETE} ->
		      [{mod_op=`REPLACE;
			mod_value={attr_type="useraccountcontrol";
				   attr_vals=["66082"]}}]
		  | modop -> [modop])
	       (List.filter (* filter out empty ADD/REPLACE ops, and delete
			       ops with match the password filters *)
		  (function
		       {mod_op=op;mod_value={attr_vals=[]}}
			 when (op = `ADD || op = `REPLACE) -> false
		     | {mod_op=`DELETE;mod_value={attr_vals=[]}} -> true
		     | {mod_op=`DELETE;mod_value={attr_vals=vals}} ->
			 if (List.for_all
			       (fun pwd ->
				  (List.exists
				     (fun rx -> Pcre.pmatch ~rex:rx pwd)
				     !cfg.cf_userpassword_filter))
			       vals)
			 then false
			 else true
		     | _ -> true)
		  (List.rev_map
		     (* filter out adds and replaces of passwords
			matching a password filter. For deletes, we
			will filter out in the next filter up the
			chain the whole op. This is because an empty
			delete has special meaning, so we can't filter
			it out empyt deletes unilaterally, like we do
			with other operation types. *)
		     (function 
			  ({mod_op=op;
			    mod_value=({attr_vals=attr_vals} as mod_val)} as mod_op)
			    when op = `ADD || op = `REPLACE ->			      
			      {mod_op with mod_value=
				  {mod_val with attr_vals=
				      (List.filter
					 (fun pwd ->
					    (not
					       (List.exists
						  (fun rx -> Pcre.pmatch ~rex:rx pwd)
						  !cfg.cf_userpassword_filter)))
					 attr_vals)}}
			| mod_op -> mod_op)
		     (List.rev_map (* change replace password [] into delete password *)
			(function
			     ({mod_op=`REPLACE;mod_value={attr_vals=[]}} as mod_op) ->
			       {mod_op with mod_op=`DELETE}
			   | mod_op -> mod_op)
			userpassword_mods))))))
  in

  (* deal with generic attribute mappings as defined in the
     configuration *)
  let map_attributes mods =
    (List.flatten
       (List.rev_map
	  (fun modop ->
	     let {mod_op=optype;
		  mod_value={attr_type=attr_name;
			     attr_vals=attr_vals}} = modop
	     in
	       try
 		 let mapping = 
		   (List.find 
		      (fun {attribute_to_match=attr_rex} -> 
			 Pcre.pmatch ~rex:attr_rex attr_name)
		      !cfg.cf_attribute_mappings)
		 in
		   [modop; (* add in mappings *)
		    {modop with 
		       mod_value={modop.mod_value with 
				    attr_type=mapping.attribute_to_emit}}]
	       with Not_found -> [modop])
	  mods))
  in

  (* remove attrs not in schema, and enforce single value *)
  let strip_not_allowed mods =
    (List.filter
       (fun {mod_value={attr_type=attr_name;attr_vals=attr_vals}} ->
	  (try ignore (attrToOid schema (Lcstring.of_string attr_name));true
	   with Invalid_attribute _ -> false) &&
	    (try 
	       match getAttr schema (Lcstring.of_string attr_name) with
		   {at_single_value=true} -> 
		     if List.length attr_vals <= 1 then
		       true
		     else 
		       raise (Single_value attr_name)
		 | _ -> true
	     with Invalid_attribute _ -> false) &&
	    (not (lc attr_name = !cfg.cf_real_replica_rdn)))
       mods)
  in
  let strip_not_allowed_objectclass mods =
    (List.filter
       (function {mod_value={attr_type=attr_name;attr_vals=[]}} -> false
	  | _ -> true)
       (List.rev_map
	  (fun ({mod_value={attr_type=attr_name;attr_vals=attr_vals}} as mod_op) ->
	     match lc attr_name with
		 "objectclass" ->
		   {mod_op with mod_value=
		       {attr_type=attr_name;
			attr_vals=
			   (List.filter
			      (fun oc ->
				 try ignore (ocToOid schema (Lcstring.of_string oc));true
				 with Invalid_objectclass _ -> false)
			      attr_vals)}}
	       | _ -> mod_op)
	  mods))
  in
  let ad_mods =    
    (strip_not_allowed_objectclass
       (strip_not_allowed
	  (map_attributes
	     (if !cfg.cf_replica_type = Active_directory then
		translate_userpassword_mods_for_ad mods
	      else 
		mods))))
  in
    {mod_dn=ad_dn;modification=ad_mods}		     

let protocol_modop_to_funclient_modop {modification=mods} =
  List.rev_map
    (fun {mod_op=op;mod_value={attr_type=attr;attr_vals=vals}} ->
       (op, attr, vals))
    mods

let add_ids_to_group cfg connection dn {modification=mods} = 
  let group_membership_mods = 
    List.filter
      (fun {mod_value={attr_type=name}} ->
	 (lc name) = (lc !cfg.cf_imaginary_group_join_attr))
      mods
  in
  let translated_group_mods = 
    List.filter
      (fun {mod_op=op;mod_value={attr_vals=vals}} -> (not (vals = [])) || (op = `DELETE))
      (List.rev_map
	 (fun ({mod_value={attr_vals=vals}} as mod_op) ->
	    let translated_vals = 
	      List.fold_left
		(fun tlst valu ->
		   try (Tdir_utils.find_object_dn cfg valu) :: tlst
		   with Tdir_utils.Util_failure _ -> tlst)
		[] vals
	    in
	      {mod_op with 
		 mod_value={attr_type=(!cfg.cf_real_group_join_attr);
			    attr_vals=translated_vals}})
	 group_membership_mods)      
  in
    match translated_group_mods with
	[] -> syslog !cfg.cf_syslog `LOG_INFO "no modifications to group membership"
      | mods -> 
	  syslog !cfg.cf_syslog `LOG_INFO ("applying modifications to group membership to dn: " ^ dn);
	  try
	    let obj =
	      List.hd
		(!cfg.cf_read_con#search 
		   ~scope:`BASE ~attrs:["objectclass";"memberof"]
		   ~base:dn "objectclass=*")
	    in
	    let lcl l = List.rev_map String.lowercase l in
	      if List.mem "group" (lcl (obj#get_value "objectclass")) then
		(syslog !cfg.cf_syslog `LOG_INFO "object we are modifying is a group";
		 connection.con_adcon#modify dn
		   (protocol_modop_to_funclient_modop {mod_dn=dn;modification=mods}))
	      else if List.mem "user" (lcl (obj#get_value "objectclass")) then
		(syslog !cfg.cf_syslog `LOG_INFO "object we are modifying is a user";
		(* In this case, we are modifying a user object. In
		   AD, group memberships must be added/removed to/from
		   the group object. So in the case of an add/delete
		   the values of the modification to the user contain
		   the dns of the groups which we must modify. We will
		   add or remove the user's own dn from all the groups
		   which she specified. In the case of a replace,
		   things get very complex, because we must compute
		   the difference between the old and new arrangment,
		   and make AD reflect that difference *)
		let rec group_updates_to_user mods = 
		  List.iter
		    (fun {mod_op=op;mod_value={attr_vals=values}} ->
		       match op with
			   `REPLACE ->
			     let present_memberships = 
			       List.rev_map 
				 (fun dn -> Ldap_dn.of_string (String.lowercase dn))
				 (try obj#get_value "memberof" with _ -> [])
			     in
			     let replacement_memberships = 
			       List.rev_map
 				 (fun dn -> Ldap_dn.of_string (String.lowercase dn))
				 values
			     in
			     let removed = 
			       List.rev_map Ldap_dn.to_string
				 (List.filter
				    (fun present -> 
				       not (List.mem present replacement_memberships))
				    present_memberships)
			     in
			     let added = 
			       List.rev_map Ldap_dn.to_string
				 (List.filter
				    (fun replacement -> 
				       not (List.mem replacement present_memberships))
				    replacement_memberships)
			     in
			       (match removed with
				    [] -> ()
				  | removed ->
				      group_updates_to_user
					[{mod_op=`DELETE;
					  mod_value=
					     {attr_type=(!cfg.cf_real_group_join_attr);
					      attr_vals=removed}}]);
			       (match added with
				    [] -> ()
				  | added ->
				      group_updates_to_user
					[{mod_op=`ADD;
					  mod_value=
					     {attr_type=(!cfg.cf_real_group_join_attr);
					      attr_vals=added}}])
			 | `DELETE when values = [] ->
			     let present_memberships = 
			       (try obj#get_value "memberof"
				with _ -> []) 
			     in
			       (match present_memberships with
				    [] -> ()
				  | to_remove ->
				      group_updates_to_user
					[{mod_op=`DELETE;
					  mod_value=
					     {attr_type=(!cfg.cf_real_group_join_attr);
					      attr_vals=to_remove}}])
			 | _ -> 
			     List.iter
			       (fun groupdn ->
				  try
				    if (List.for_all 
					  (fun rex -> Pcre.pmatch ~rex:rex groupdn) 
					  !cfg.cf_target_group_must_match) &&
				        (not (List.exists
						(fun rex -> Pcre.pmatch ~rex:rex groupdn)
						!cfg.cf_target_group_must_not_match))
				    then
				      (syslog 
					 !cfg.cf_syslog `LOG_INFO 
					 (sprintf "modifying %s" groupdn);
				       connection.con_adcon#modify groupdn 
					 [(op, !cfg.cf_real_group_join_attr, [dn])])
				    else
				      syslog
					!cfg.cf_syslog `LOG_INFO 
					(sprintf "group does not meet criteria: %s" groupdn)
				  with (LDAP_Failure _) as exn ->
				    syslog !cfg.cf_syslog `LOG_ERR
				      (ldap_strerror "failed to update group membership" exn))
			       values)
		    mods
		in
		  group_updates_to_user mods)
	  with (LDAP_Failure _) as exn ->
	    syslog !cfg.cf_syslog `LOG_ERR ("could not fetch user " ^ (ldap_strerror "" exn))
	      
let modify cfg {connections=connections;ad_schema=schema} conn_id message =
  match message with
      {messageID=id;
       protocolOp=Modify_request mod_op} ->
	if (List.exists (* make sure the dn meets the user's criteria *)
	      (fun rx -> Pcre.pmatch ~rex:rx mod_op.mod_dn) 
	      !cfg.cf_dn_must_not_match) || 
	   (not 
	      (List.exists
		 (fun rx -> Pcre.pmatch ~rex:rx mod_op.mod_dn)
		 !cfg.cf_dn_must_match)) ||
	   (not (Tdir_utils.match_object cfg mod_op.mod_dn))
	then begin
	  syslog !cfg.cf_syslog `LOG_DEBUG "dn does not meet criteria, not doing modify";
	  mod_res message
	end
	else begin
	  try
	    let connection = Tdir_utils.connection cfg connections conn_id in	   	   
	    let mods = translate_modop cfg schema mod_op in
	     (match mods.modification with
		  [] -> 
		    (if !cfg.cf_replica_type = Active_directory then
		       add_ids_to_group cfg connection mods.mod_dn mod_op);
		    mod_res message
		| _ ->
		    connection.con_adcon#modify mods.mod_dn 
		      (protocol_modop_to_funclient_modop mods);
		    (if !cfg.cf_replica_type = Active_directory then
		       add_ids_to_group cfg connection mods.mod_dn mod_op);
		    mod_res message)
	  with 
	      Single_value attr -> 
		(mod_res ~result_code:`OBJECT_CLASS_VIOLATION 
		   ~error_msg:(attr ^ " is single valued, you have specified multiple values")
		   message)
	    | LDAP_Failure (res, err, {ext_matched_dn=mdn;ext_referral=referral}) -> 
		(mod_res ~result_code:res ~error_msg:err ~matched_dn:mdn 
		   ~referral:referral message)
	    | Tdir_utils.Util_failure msg -> 
		mod_res ~result_code:`INVALID_DN_SYNTAX ~error_msg:msg message
	    | LDAP_Decoder e -> 
		mod_res ~result_code:`OTHER ~error_msg:("AD Protocol error: " ^ e) message
	    | exn -> 
		mod_res message ~result_code:`OTHER 
		  ~error_msg:("unhandled exception: " ^ (Printexc.to_string exn))
	end
    | _ -> mod_res ~result_code:`PROTOCOL_ERROR ~error_msg:"invalid modify request" message
